# =============================================================================
# MSP430 Access Library
# =============================================================================
import std/options
import private/i2c

type
  Msp430Obj = object
    dev: I2cdev
  Msp430* = ref Msp430Obj
  Pol* {.pure.} = enum
    Rising = "Rising"
    Falling = "Falling"

const
  DI_NUMS = 5
  FILTERED_DI = 0x06'u8
  DI_POLARITY = 0x0b
  CounterCtrl = 0x10'u8
  CounterCompare0 = 0x13'u8
  CounterVal0 = 0x18'u8

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
func counter_compare(idx: int): uint8 =
  result = CounterCompare0 + idx.uint8

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
func counter_value(idx: int): uint8 =
  result = CounterVal0 + idx.uint8

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc `=destroy`(self: var Msp430obj) =
  self.dev = nil

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc newMsp430*(bus: int, address: uint8, debug: bool = false): Msp430 =
  let i2cdev = newI2cdev(bus, address, debug)
  result = new Msp430
  result.dev = i2cdev

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc write_register(self: Msp430, address: uint8, val: uint8): bool =
  let buf: array[2, uint8] = [address, val]
  result = self.dev.write(buf)

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc read_register*(self: Msp430, address: uint8): Option[uint8] =
  let reg = [address]
  let val = self.dev.write_read(reg, 1)
  if val.len == 1:
    result = some(val[0].uint8)

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc read_registers*(self: Msp430, address: uint8, count: int): Option[seq[uint8]] =
  let reg = [address]
  let values = self.dev.write_read(reg, count)
  if values.len == count:
    var res = newSeqOfCap[uint8](count)
    for val in values:
      res.add(val.uint8)
    result = some(res)

# -------------------------------------------------------------------
# API: Get DI Status Register
# -------------------------------------------------------------------
proc get_di_status*(self: Msp430): Option[seq[int]] =
  let stat_opt = self.read_register(FILTERED_DI)
  if stat_opt.isNone:
    return
  let stat = stat_opt.get()
  var res = newSeqOfCap[int](DI_NUMS)
  for pos in 0 ..< DI_NUMS:
    let val = if (stat and (1'u8 shl pos)) != 0: 1 else: 0
    res.add(val)
  result = some(res)

# -------------------------------------------------------------------
# API: Get DI Polarity Register
# -------------------------------------------------------------------
proc get_di_polarity*(self: Msp430): Option[seq[Pol]] =
  let stat_opt = self.read_register(DI_POLARITY)
  if stat_opt.isNone:
    return
  let stat = stat_opt.get()
  var res = newSeqOfCap[Pol](DI_NUMS)
  for pos in 0 ..< DI_NUMS:
    let pol = if (stat and (1'u8 shl pos)) != 0: Pol.Falling else: Pol.Rising
    res.add(pol)
  result = some(res)

# -------------------------------------------------------------------
# API: Set DI Polarity Register
# -------------------------------------------------------------------
proc set_di_polarity*(self: Msp430, idx: int, pol: Pol): bool =
  if idx < 0 or idx >= DI_NUMS:
    return
  let reg_opt = self.read_register(DI_POLARITY)
  if reg_opt.isNone:
    return
  var reg = case pol
    of Pol.Rising: reg_opt.get() and (not (1'u8 shl idx))
    of Pol.Falling: reg_opt.get() or (1'u8 shl idx)
  result = self.write_register(DI_POLARITY, reg)

# -------------------------------------------------------------------
# API: Get Counter Control Register
# -------------------------------------------------------------------
proc get_counter_control*(self: Msp430): Option[seq[bool]] =
  let stat_opt = self.read_register(CounterCtrl)
  if stat_opt.isNone:
    return
  let stat = stat_opt.get()
  var res = newSeqOfCap[bool](DI_NUMS)
  for pos in 0 ..< DI_NUMS:
    let onoff = (stat and (1'u8 shl pos)) != 0
    res.add(onoff)
  result = some(res)

# -------------------------------------------------------------------
# API: Get Counter Control Register (selected DI only)
# -------------------------------------------------------------------
proc get_counter_onoff*(self: Msp430, idx: int): Option[bool] =
  if idx < 0 or idx >= DI_NUMS:
    return
  let stat_opt = self.read_register(CounterCtrl)
  if stat_opt.isNone:
    return
  let onoff = (stat_opt.get() and (1'u8 shl idx)) != 0
  result = some(onoff)

# -------------------------------------------------------------------
# API: Enable/Disable Counter
# -------------------------------------------------------------------
proc onoff_counter*(self: Msp430, idx: int, onoff: bool): bool =
  if idx < 0 or idx >= DI_NUMS:
    return false
  let stat_opt = self.read_register(CounterCtrl)
  if stat_opt.isNone:
    return false
  let stat_current = stat_opt.get()
  var stat_new: uint8
  let mask = 1'u8 shl idx
  if onoff:
    stat_new = stat_current or mask
  else:
    stat_new = stat_current and (not mask)
  if stat_new == stat_current:
    return true
  result = self.write_register(CounterCtrl, stat_new)

# -------------------------------------------------------------------
# API: Get Counter Value
# -------------------------------------------------------------------
proc get_counter_val*(self: Msp430, idx: int): Option[uint8] =
  if idx < 0 or idx >= DI_NUMS:
    return
  let val_opt = self.read_register(counter_value(idx))
  if val_opt.isNone:
    return
  result = some(val_opt.get())


when isMainModule:
  import strformat

  let msp430 = newMsp430(1, 0x4f'u8)
  let ctrl_opt = msp430.get_counter_control()
  if ctrl_opt.isSome:
    echo &"Counter Control: {ctrl_opt.get}"
    for idx, onoff in ctrl_opt.get.pairs:
      echo &"* Counter[{idx}] ON/OFF: {onoff}"
      if not onoff:
        discard msp430.onoff_counter(idx, true)
  for idx in 0..<5:
    let val_opt = msp430.get_counter_val(idx)
    if val_opt.isSome:
      echo &"* Counter[{idx}] -> value: {val_opt.get}"
