# =============================================================================
# I2C Access Library
# =============================================================================
import std/posix
import std/sequtils
import std/strformat

type
  I2cdevObj = object
    fd: File
    address: uint8
    debug: bool
    opened: bool
  I2cdev* = ref I2cdevObj
  I2c_msg {.importc: "struct i2c_msg", header: "<linux/i2c.h>".} = object
    `addr`: cushort  # slave address
    flags: cushort
    len: cushort
    buf: cstring
  I2c_rdwr_ioctl_data {.importc: "struct i2c_rdwr_ioctl_data",
      header: "<linux/i2c-dev.h>".} = object
    msgs: ptr I2c_msg #  pointers to i2c_msgs
    nmsgs: cuint      #  number of i2c_msgs

const
  I2C_M_RD = 0x0001  # read data, from slave to master
  I2C_RDWR = 0x0707

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc `=destroy`(self: var I2cdevObj) =
  if self.opened:
    self.fd.close()
    self.opened = false

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc newI2cdev*(bus: int, address: uint8, debug: bool = false): I2cdev =
  let devname = &"/dev/i2c-{bus}"
  var fd: File
  try:
    fd = open(devname, fmReadWrite)
  except:
    return
  result = new I2cdev
  result.fd = fd
  result.address = address
  result.opened = true
  result.debug = debug

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc write_read*(self: I2cdev, writebuf: openArray[char|uint8], readlen: int):
    seq[char] =
  var packets: I2c_rdwr_ioctl_data
  var msgs: array[2, I2c_msg]
  var rd_buf: array[256, char]

  # Setting up the register write
  msgs[0].`addr` = self.address.uint16
  msgs[0].flags = 0
  msgs[0].len = writebuf.len.uint16
  msgs[0].buf = cast[cstring](unsafeAddr writebuf)
  # Setting up the read
  msgs[1].`addr` = self.address.uint16
  msgs[1].flags = I2c_M_RD
  msgs[1].len = readlen.uint16
  msgs[1].buf = cast[cstring](addr rd_buf)
  packets.msgs = addr msgs[0]
  packets.nmsgs = 2

  let res = posix.ioctl(self.fd.getFileHandle, I2C_RDWR, addr packets)
  if res < 0:
    return @[]
  result = rd_buf.toSeq()
  result.setLen(readlen)

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc write*(self: I2cdev, writebuf: openArray[char|uint8]): bool =
  var packets: I2c_rdwr_ioctl_data
  var msgs: array[1, I2c_msg]

  # Setting up the register write
  msgs[0].`addr` = self.address.uint16
  msgs[0].flags = 0
  msgs[0].len = writebuf.len.uint16
  msgs[0].buf = cast[cstring](unsafeAddr writebuf)
  packets.msgs = addr msgs[0]
  packets.nmsgs = 1

  let res = posix.ioctl(self.fd.getFileHandle, I2C_RDWR, addr packets)
  if res < 0:
    result = false
  else:
    result = true

# -------------------------------------------------------------------
#
# -------------------------------------------------------------------
proc read*(self: I2cdev, readlen: int): seq[char] =
  var packets: I2c_rdwr_ioctl_data
  var msgs: array[1, I2c_msg]
  var rd_buf: array[256, char]

  # Setting up the register write
  msgs[0].`addr` = self.address.uint16
  msgs[0].flags = I2C_M_RD
  msgs[0].len = readlen.uint16
  msgs[0].buf = cast[cstring](addr rd_buf)
  packets.msgs = addr msgs[0]
  packets.nmsgs = 1

  let res = posix.ioctl(self.fd.getFileHandle, I2C_RDWR, addr packets)
  if res < 0:
    result = @[]
  else:
    result = rd_buf.toSeq()
    result.setLen(readlen)


when isMainModule:
  proc bcd2bin(bcd: char): int =
    result = (((bcd.uint8 and 0xf0) shr 4) * 10 + (bcd.uint8 and 0x0f)).int

  var i2c = newI2cdev(1, 0x32)
  if i2c.isNil:
    quit("open i2c failed")
  var wbuf: seq[uint8] = @[0'u8]

  var buf = i2c.write_read(wbuf, 0x0d + 1)
  echo fmt"buf lenght: {buf.len}"
  if buf.len > 0:
    try:
      let year = bcd2bin(buf[6]) + 2000
      let month = bcd2bin(buf[5])
      let day = bcd2bin(buf[4])
      let hour = bcd2bin(buf[2])
      let minute = bcd2bin(buf[1])
      let second = bcd2bin(buf[0])
      echo fmt"{year}/{month:02d}/{day:02d} {hour:02d}:{minute:02d}:{second:02d} [UTC]"
    except:
      echo "length error"
